var InspectStartPlate = function()
{
	var inspectstartplate = {};
	var barcodesByName = [];
	
	var EXTERNAL_DNA_NAME = 'External.d';
	var debug = 0;
	
	var currentSelected;
	var names = [];
	var wells = [];
	var protocolMap = {};
	var kitNum = 0;
	var selectedKitName;

	var graphics;
	var pen;
	
	var quantitiesAreValid = false;
	var plateNameIsValid = false;
	var subtypeDna;
	var subtypeDnaNormalized;
	var MINIMAL_DNA_VOLUME = 1.0; //µl
	var THIN_SPACE = '&#x2005;';
	
	var jsonRowIndex = {"A": 0, "B": 1, "C": 2, "D": 3, "E": 4, "F": 5, "G": 6, "H": 7};
	var jsonColIndex = {"1": 0, "2": 1, "3": 2, "4": 3, "5": 4, "6": 5, "7": 6, "8": 7, "9": 8, "10": 9, "11": 10, "12": 11};
	var jsonRowLetter = {0: "A", 1: "B", 2: "C", 3: "D", 4: "E", 5: "F", 6: "G", 7: "H"};
	var jsonColNumber = {0: "1", 1: "2", 2: "3", 3: "4", 4: "5", 5: "6", 6: "7", 7: "8", 8: "9", 9: "10", 10: "11", 11: "12"};

	// Page initialization
	inspectstartplate.initPage = function()
	{
		// Step 1
		Events.addEventHandler('step-1', 'wizard-validate', inspectstartplate.validateStep1);

		// Step 2
		Events.addEventHandler('step-2', 'wizard-initialize', inspectstartplate.initializeStep2);
		Events.addEventHandler('step-2', 'wizard-validate', inspectstartplate.validateStep2);

/*
		Events.addEventHandler('plateName', 'change', inspectstartplate.plateNameOnChange);
		Events.addEventHandler('pool_schema', 'change', inspectstartplate.poolSchemaOnChange);
		Events.addEventHandler('barcode_variant', 'change', inspectstartplate.barcodeVariantOnChange);
*/
		Events.addEventHandler('min_vol_dna', 'change', inspectstartplate.minVolDnaOnChange);
		Events.addEventHandler('min_vol_dna', 'keypress', Events.numberOnly);
		
		//Events.addEventHandler('warning_quantity', 'change', inspectstartplate.warningLevelOnChange);
		Events.addEventHandler('warning_quality_score', 'change', inspectstartplate.warningLevelOnChange);
		//Events.addEventHandler('warning_quantity', 'keypress', Events.numberOnly);
		Events.addEventHandler('warning_quality_score', 'keypress', Events.numberOnly);
		
		Events.addEventHandler('location', 'click', inspectstartplate.toggleInfo);
		Events.addEventHandler('quantity', 'click', inspectstartplate.toggleInfo);
		Events.addEventHandler('quality-score', 'click', inspectstartplate.toggleInfo);
		Events.addEventHandler('qubitconc', 'click', inspectstartplate.toggleInfo);
		Events.addEventHandler('volumes', 'click', inspectstartplate.toggleInfo);
/*
		Events.addEventHandler('qiacube-date', 'click', inspectstartplate.toggleInfo);
*/
		
		Events.addEventHandler('pool-row', 'click', inspectstartplate.togglePool);
		Events.addEventHandler('pool-row', 'mouseover', inspectstartplate.highlightPool);
		Events.addEventHandler('pool-row', 'mouseout', inspectstartplate.highlightPool);

		Buttons.addClickHandler('downloadLibPrepFile', inspectstartplate.downloadLibPrepFile);

		Events.addEventHandler('iconSpecialSelect', 'click', inspectstartplate.toggleSpecialSelect);
		Events.addEventHandler('plate', 'mouseup', inspectstartplate.contextEvent);
		Events.addEventHandler('plate', 'contextmenu', inspectstartplate.contextEvent);
		Events.addEventHandler('mnuComment', 'click', inspectstartplate.commentSelected);
		Events.addEventHandler('mnuCaseSummary', 'click', inspectstartplate.showCaseSummary);

		Events.addEventHandler('flaggedDna', 'click', inspectstartplate.showFlaggedDna);

		Events.addEventHandler('listview', 'click', inspectstartplate.viewProtocol);
		Events.addEventHandler('plateview', 'click', inspectstartplate.viewProtocol);

		// Navigation
		Buttons.addClickHandler('gocancel', Wizard.cancelWizard);
		Buttons.addClickHandler('gonext', Wizard.goNextOnClick);
		Buttons.addClickHandler('gocreate', inspectstartplate.createProtocol);

/*		
		// Final registration
		Events.addEventHandler('wizard', 'wizard-submit', inspectstartplate.submit);
*/

		var isNeoPrep = Data.int('page-data', 'is-neoprep');

		var url = '../Dna.servlet?ID='+App.getSessionId();
		url += '&cmd=GetStartPlates';
		url += '&bioPlateType=' + (isNeoPrep ? 'NEOPREP' : 'DNA');
		url += '&bioPlatePrefix=' + 'LP';
		url += '&orderDesc=true';
		Wizard.showLoadingAnimation('Loading '+(isNeoPrep ? 'NeoPrep' : 'DNA') + ' plates');
		Wizard.asyncJsonRequest(url, inspectstartplate.initializeStep1);
	}

	inspectstartplate.initializeStep1 = function(response)
	{
		var frm = document.forms['meludi'];
		var jsonPlates = response.bioplates;

		var startPlate = frm.startPlate;
		if (jsonPlates.length > 0)
		{
			for (var i=0; i < jsonPlates.length; i++)
			{
				var plate = jsonPlates[i];
				var name = (i+1) + ': ' + Strings.encodeTags(plate.name);
				if (plate.numMeludiItems || plate.numExtraItems)
				{
					name += ' -- (';
					name += plate.numMeludiItems;
					name += ' + ' + plate.numExtraItems;
					name += ')';
				}
				if (plate.registrationDate)
				{
					name += ' -- [' + inspectstartplate.asDate(plate.registrationDate) + ']';
				}
				var selected = (i == 0);
				var option = new Option(name, plate.id, selected, selected);
				option.plate = plate;
				startPlate.options[startPlate.length] = option;
			}
			startPlate.options[0].selected = true;
		}
		else
		{
			Wizard.setFatalError('No start plates available.');
			return;
		}
		Doc.show('step-1');
		Doc.show('gocancel');
		Doc.show('gonext');
		Doc.hide('gocreate');
		Doc.hide('extractSourceItems');
	}

	inspectstartplate.validateStep1 = function(event)
	{
		var frm = document.forms['meludi'];
		var startPlate = null;
		for (var i = 0; i < frm.startPlate.length; i++)
		{
			if (frm.startPlate[i].selected)
			{
				startPlate = frm.startPlate[i];
			}
		}
		if (startPlate == null)
		{
			event.preventDefault();
		}
	}
	
	inspectstartplate.initializeStep2 = function(event)
	{
		var frm = document.forms['meludi'];
		var isNeoPrep = Data.int('page-data', 'is-neoprep');

		var jsonPlate = null;
		for (var i = 0; i < frm.startPlate.length; i++)
		{
			if (frm.startPlate[i].selected)
			{
				jsonPlate = frm.startPlate[i];
			}
		}
		if (jsonPlate == null)
		{
			event.preventDefault();
		}
		var startPlate = jsonPlate.plate;

		frm.startPlate.disabled = true;

		Wizard.setCurrentStep(2);
		Doc.show('step-2');
		Doc.show('gocancel');
		Doc.hide('gonext');

		Plate.name = startPlate.name;
		var protocolName = startPlate.protocolName;
		var kitName = startPlate.kitName;
		var comment = startPlate.comments;
		var wellsInfo = startPlate.wells;

		frm.plateName.value = Strings.encodeTags(Plate.name);
		frm.plateName.disabled = true;
		plateNameIsValid = true;
		Wizard.setInputStatus('plateName', 'valid');

		frm.libraryPreparationProtocol.value = protocolName;
		frm.libraryPreparationProtocol.disabled = true;
		//
		frm.tsLibPrepKit.value = kitName;
		frm.tsLibPrepKit.disabled = true;
		frm.hiddenLibPrepKitId.value = startPlate.kitId;

		frm.comments.value = comment;
		frm.comments.disabled = true;
		
		// Place pre-selected DNA on plate
		inspectstartplate.quantityOnChange();
		inspectstartplate.warningLevelOnChange();
		
		// Init plate, pool schema and barcode variants
		graphics = new jsGraphics(Doc.element('canvas'));
		pen = new jsPen(new jsColor('#2288AA'), 2);
		var columns = isNeoPrep ? 2 : 12;
		var rows = 8;
		// Disable pool schema menu
		frm.pool_schema.disabled = true;
		var schema = PoolSchema.initList(frm.pool_schema, null, isNeoPrep ? 'neoprep' : 'manual');
		Plate.init(rows, columns, schema, WellPainter);
		Events.sendChangeEvent(frm.pool_schema);
		// Set default for showing info for plate well
		inspectstartplate.setDefaultInfo('location');
		inspectstartplate.setDefaultInfo('volumes');
		inspectstartplate.setDefaultInfo('quality-score');
		inspectstartplate.setDefaultInfo('qubitconc');

		// Fill plate wells with well data for selected start plate
		for (var column = 0; column < 12; column++)
		{
			for (var row = 0; row < 8; row++)
			{
				var wellStr = inspectstartplate.wellRowColumnToWellStr(row, column);
				// Get well info from start plate
				var wellInfo = wellsInfo[wellStr];
				if (wellInfo)
				{				
					var well = Plate.getWell(row, column);
					var parentName = wellInfo.extract.origName;
					well.extract = wellInfo.extract;
					var dnaId = wellInfo.extract.id;
					// Create a new DNA object
					//var dna = Dna.createByName(name);
					var dnaExt = '.fpa';
					if (column >= 6)
					{
						dnaExt = '.fpb';
					}
					var dna = Dna.createByParentName(parentName, parentName + dnaExt, dnaId);
					Dna.unflag(dna);
					well.setExtract(dna);
					well.selected = false;
				}
			}
		}	
		Plate.checkReplicates();
		Plate.paint(Plate.getWells());
		inspectstartplate.updateNumFlaggedDna();
		names = [];
		//
		Doc.show('step-3');
		Doc.show('gocancel');
		Doc.show('gocreate');
		Wizard.setNoConfirmOnFirstStep(false);
		Wizard.keepSessionAlive();
	}

	inspectstartplate.wellStrToRow = function(wellStr)
	{
		// Get first letter
		var rowStr = wellStr.substring(0,1);
		var row = jsonRowIndex[rowStr];
		return row;
	}

	inspectstartplate.wellStrToCol = function(wellStr)
	{
		// Get column string;
		var colStr = wellStr.substring(1,3);
		var col = jsonColIndex[colStr];
		return col;
	}

	inspectstartplate.wellRowColumnToWellStr = function(row, column)
	{
		// Get well string (row letter + column number, 1-based)
		var wellStr = jsonRowLetter[row] + jsonColNumber[column];
		return wellStr;
	}

	inspectstartplate.plateNameOnChange = function(event)
	{
		var frm = document.forms['meludi'];
		plateNameIsValid = false;
		
		var name = frm.plateName.value;
		if (!name || name == '')
		{
			Wizard.setInputStatus('plateName', 'invalid', 'Missing name');
			return;
		}
		if (name.indexOf("LP") != 0)
		{
			Wizard.setInputStatus('plateName', 'invalid', 'Plate name must start with prefix "LP"');
			return;
		}
		if (name.length != 7)
		{
			Wizard.setInputStatus('plateName', 'invalid', 'Plate name must consist of "LP" + 5 digits');
			return;
		}
		
		// Check if name exists
		var url = '../Extraction.servlet?ID='+App.getSessionId();
		url += '&cmd=CheckIfStartPlateNameUsed';
		url += '&startPlateName=' + encodeURIComponent(name);
		
		Wizard.showLoadingAnimation('Loading information about plate...');
		Wizard.asyncJsonRequest(url, inspectstartplate.plateInfoLoaded);
	}

	inspectstartplate.plateInfoLoaded = function(response)
	{
		var frm = document.forms['meludi'];	
		var plateInfo = response.startplate;
		if (plateInfo)
		{
			var plateName = plateInfo.name;
			var plateId = plateInfo.id;

			if (plateId)
			{
				plateNameIsValid = false;
				Wizard.setInputStatus('plateName', 'invalid', 'Plate already exists with name "' + plateName + '" (ID = ' + plateId + ')');
				return;
			}
		}
		plateNameIsValid = true;
		Plate.name = plateInfo.name;
		
		Wizard.setInputStatus('plateName', 'valid');
	}

	inspectstartplate.quantityOnChange = function(event)
	{
		var frm = document.forms['meludi'];
		quantitiesAreValid = false;

/*		
		var qRegular = parseFloat(frm.quantity_regular.value);
		if (!(qRegular > 0))
		{
			Wizard.setInputStatus('quantities', 'invalid', 'Amount must be ≥ 0.');
			return;
		}
*/
/*
		var qQc = parseFloat(frm.quantity_qc.value);
		if (!(qQc > qRegular))
		{
			Wizard.setInputStatus('quantities', 'invalid', 'Amount QC must be &gt; ' + qRegular + '.');
			return;
		}
*/
		
		quantitiesAreValid = true;
/*
		QUANTITY_REGULAR = qRegular;
*/
/*
		QUANTITY_QC = qQc;
*/

/*		
		Wizard.setInputStatus('quantities', 'valid');
*/
		if (event) Plate.paint(Plate.getWells());
	}

	inspectstartplate.minVolDnaOnChange = function(event)
	{
		var frm = document.forms['meludi'];
		quantitiesAreValid = false;
		
		var minVolDna = parseFloat(frm.min_vol_dna.value);
		if (!(minVolDna > 0))
		{
			Wizard.setInputStatus('min_vol_dna', 'invalid', 'Amount must be ≥ 0.');
			return;
		}
		
		quantitiesAreValid = true;
		MINIMAL_DNA_VOLUME = minVolDna;
		
		Wizard.setInputStatus('min_vol_dna', 'valid');
		if (event) Plate.paint(Plate.getWells());
	}

	inspectstartplate.warningLevelOnChange = function(event)
	{
		var frm = document.forms['meludi'];

/*
		var wQuantity = parseFloat(frm.warning_quantity.value);
		if (!(wQuantity > 0))
		{
			Wizard.setInputStatus('warnings', 'invalid', 'Remaining quantity must be ≥ 0.');
			return;
		}
*/
		var wQualityScore = parseFloat(frm.warning_quality_score.value);

/*		
		LOW_QUANTITY_WARNING_LIMIT = wQuantity;
*/
		QUALITY_SCORE_WARNING_LIMIT = wQualityScore;
		
		Wizard.setInputStatus('warnings', 'valid');
		if (event) Plate.paint(Plate.getWells());
	}
	
	inspectstartplate.poolSchemaOnChange = function()
	{
		var frm = document.forms['meludi'];
		var schema = PoolSchema.getById(frm.pool_schema.value);
		Plate.setPoolSchema(schema);
		
		var isNeoPrep = Data.int('page-data', 'is-neoprep');
		if (isNeoPrep)
		{
			PoolSchema.buildPoolTableRow(schema, Plate.columns, true, '<td class="barcode-fill barcode-left"></td>', '<th></th><td class="barcode-fill barcode-right"></td>');
			PoolSchema.initVariantList(frm.barcode_variant, schema);
			Events.sendChangeEvent(frm.barcode_variant);
		}
		else
		{
			PoolSchema.buildPoolTableRow(schema, Plate.columns, true);
			Plate.paint(Plate.getWells());
		}
	}
	
	inspectstartplate.barcodeVariantOnChange = function()
	{
		var frm = document.forms['meludi'];
		var schema = PoolSchema.getById(frm.pool_schema.value);
		var barcodeVariant = PoolSchema.getBarcodeVariantByName(schema, frm.barcode_variant.value);
		
		inspectstartplate.assignSchemaDefinedBarcodeVariant(schema, barcodeVariant);
		WellPainter.barcodeVariant = barcodeVariant;
		Plate.paint(Plate.getWells());
	}

	inspectstartplate.assignSchemaDefinedBarcodeVariant = function(schema, barcodeVariant)
	{
		var wells = Plate.getWells();
		for (var i = 0; i < wells.length; i++)
		{
			var well = wells[i];

			var indexSet = barcodeVariant ? barcodeVariant.indexSets[well.column] : null;
			var barcode = null;
			if (indexSet)
			{
				var barcodeName = indexSet.barcodes[well.row];
				barcode = barcodesByName[barcodeName];
			}
			well.barcode = barcode;
			well.defaultBarcode = barcode;
			well.duplicate = false;
		}
	}

	var currentWell;
	var lastSelectedBarcode;
	inspectstartplate.selectBarcode = function(event)
	{
		var row = Data.int(event.currentTarget, 'row');
		var column = Data.int(event.currentTarget, 'col');

		// Locate the barcode selection div so that the current
		// barcode is positioned next to the current mouse position
		// Initially set it to the right of the mouse so that the
		// center is vertically aligned
		currentWell = Plate.getWell(row, column);
		var scroll = 0;

		// Reset 'current' selection
		var menu = Doc.element('select-barcode');
		var selectAll = Doc.element('select-barcode-all');
		for (var i = 0; i <  selectAll.childNodes.length; i++)
		{
			Doc.removeClass(selectAll.childNodes[i], 'current');
		}
		menu.style.display = 'block';

		var x = event.clientX+1;
		var halfHeight = Math.floor(selectAll.offsetHeight/2)
		var y = event.clientY-halfHeight;
		var scroll = 0;
		var barcodeDiv;
		
		var useBarcode = currentWell.barcode || lastSelectedBarcode;
		if (useBarcode)
		{
			var barcodeDiv = Doc.element('bc-'+useBarcode.id);
			if (!currentWell.barcode && barcodeDiv.nextSibling) barcodeDiv = barcodeDiv.nextSibling;
			// Try to scroll the current barcode so that it's baseline is at the center of the div
			scroll = barcodeDiv.offsetTop + barcodeDiv.offsetHeight - halfHeight;
			if (scroll < 0) 
			{
				// We get a negative scroll for the first few elements, shift the
				// entire selection div down instead
				y -= scroll;
				scroll = 0;
			}
			else if (scroll > selectAll.scrollHeight - selectAll.offsetHeight)
			{
				// We get a too large scroll value for the last few elements, shift
				// the entire selection div up instead
				y -= scroll - (selectAll.scrollHeight - selectAll.offsetHeight);
				scroll = selectAll.scrollHeight - selectAll.offsetHeight;
			}
			if (currentWell.barcode)
			{
				Doc.addClass(barcodeDiv, 'current');
			}
		}
		
		// Default barcode
		if (currentWell.defaultBarcode && currentWell.defaultBarcode != currentWell.barcode)
		{
			var selectDefault = Doc.element('select-barcode-default');
			Data.set(selectDefault, 'barcode-name', currentWell.defaultBarcode.name);
			selectDefault.innerHTML = 'Default: ' + Strings.encodeTags(currentWell.defaultBarcode.name);
			Doc.show('select-barcode-default');
			Doc.show('select-barcode-default-separator');
		}
		else
		{
			Doc.hide('select-barcode-default');
			Doc.hide('select-barcode-default-separator');
		}
		
		// Position the selection div
		selectAll.scrollTop = scroll;
		menu.style.left = (x)+'px';
		menu.style.top = (y)+'px';
		event.stopPropagation();
	}

	inspectstartplate.barcodeSelected = function(event)
	{
		var frm = document.forms['meludi'];
		var schema = PoolSchema.getById(frm.pool_schema.value);
		var target = event.target;
		if (currentWell)
		{
			lastSelectedBarcode = barcodesByName[Data.get(target, 'barcode-name')];
			currentWell.barcode = lastSelectedBarcode;
			var poolNum = schema.getPoolNumForColumn(currentWell.column);
			inspectstartplate.checkDuplicateBarcode(poolNum);
			Plate.paint(Plate.getWells());
		}
		inspectstartplate.hideBarcodeSelection();
	}

	inspectstartplate.checkDuplicateBarcode = function(poolNum)
	{
		var frm = document.forms['meludi'];
		var schema = PoolSchema.getById(frm.pool_schema.value);

		var wells = schema.getWellsInPool(Plate, poolNum);
		var wellByBarcodeName = [];
		for (var i = 0; i < wells.length; i++)
		{
			var well = wells[i];
			if (well.barcode)
			{
				if (wellByBarcodeName[well.barcode.name])
				{
					well.duplicate = true;
					wellByBarcodeName[well.barcode.name].duplicate = true;
				}
				else
				{
					wellByBarcodeName[well.barcode.name] = well;
					well.duplicate = false;
				}
			}
		}
	}

	
	inspectstartplate.hideBarcodeSelection = function()
	{
		Doc.hide('select-barcode');
	}

	
	/**
		Sets default display of info data for plate wells,
		based on check box values. Assumes the id of the
		check box in question as input.
	*/
	inspectstartplate.setDefaultInfo = function(chkbox)
	{
		var frm = document.forms['meludi'];
		var show = frm[chkbox].checked;
		Doc.addOrRemoveClass('plate', 'hide-'+chkbox, !show);
	}

	inspectstartplate.toggleInfo = function(event)
	{
		var frm = document.forms['meludi'];
		var show = event.currentTarget.checked;
		Doc.addOrRemoveClass('plate', 'hide-'+event.currentTarget.id, !show);
		if (event.currentTarget.id == 'qubitconc')
		{
			Doc.addOrRemoveClass('plate', 'hide-origqubitconc', !show);
			Doc.addOrRemoveClass('plate', 'hide-calcqubitconc', !show);
		}
		// Re-paint well info to show/hide warnings for target class
		Plate.paint(Plate.getWells());
	}

	/**
		Open a popup dialog for selecting a data files.
	*/
	inspectstartplate.selectFile = function()
	{
		Dialogs.openPopup('select_file.jsp?ID='+App.getSessionId(), 'SelectFile', 600, 400);	
	}

	/**
		Parse a tab-separated data file. The columns must be in the following order:
		0: DNA name
		1: Row letter (A-H)
		2: Column number (1-12)
		3: QC flag (can be empty)
	*/
	inspectstartplate.parseDnaFile = function(data)
	{
		var lines = data.split(/[\n\r]+/);
	
		// Load information about all DNA items in a batch (will improve performance)
		var names = [];
		for (var i = 0; i < lines.length; i++)
		{
			var line = lines[i];
			if (line)
			{
				var cols = lines[i].split(/\t/);
				if (cols.length < 3) throw 'On line '+(i+1)+': Too few columns (' + cols.length + ')';
				names[names.length] = cols[0];
			}
		}
		Dna.loadInfoByNames(names);
		
		// Place DNA on the plate
		var duplicates = [];
		for (var i = 0; i < lines.length; i++)
		{
			var line = lines[i];
			if (line)
			{
				var cols = lines[i].split(/\t/);
	
				//var dna = Dna.createByName(cols[0]);
				var row =  Meludi.alphaToWell(cols[1].toUpperCase());
				var col =  parseInt(cols[2], 10)-1;

				if (col < 6)
				{				
					var well = Plate.getWell(row, col);
					if (!well) throw 'On line '+(i+1)+': Invalid plate coordinate ['+cols[1]+','+cols[2]+']';
	
					var name = cols[0];
					// Create a new DNA object
					var dna = Dna.createByName(name);
					dna.name = name + '.fpa';
					dna.qc = cols.length >= 4 && (cols[3] && cols[3] != '0');
					Dna.unflag(dna);
				
					// Check for duplicate DNA on same position (which is an error)
					var pos = 'c'+col+'r'+row;
					if (duplicates[pos])
					{
						well.addDuplicate(dna);
					}
					else
					{
						duplicates[pos] = dna.name;
						well.setExtract(dna);
					}

					// Place FPB sample in corresponding well on right side of plate
					well = Plate.getWell(row, col + 6);
					// Create a new DNA object
					dna = Dna.createByName(name);
					dna.name = name + '.fpb';
					dna.qc = cols.length >= 4 && (cols[3] && cols[3] != '0');
					Dna.unflag(dna);
					well.setExtract(dna);
				}
			}
		}
		
		inspectstartplate.updateNumFlaggedDna();
		
		// Check for replicates on the whole plate and repaint it
		Plate.checkReplicates(null, true);
		Plate.paint(Plate.getWells());
	}
	
	/**
		Open a popup dialog for manual selection of DNA.
	*/
	inspectstartplate.manualSelect = function(event)
	{
		currentSelected = Plate.getSelected();
		if (currentSelected.length == 0)
		{
			Forms.showNotification(event.currentTarget, 'Please select one or more wells were DNA should be placed.');
			return;
		}
		
		if (subtypeDna == null) subtypeDna = Meludi.getSubtypeInfo('DNA');
/*
		if (subtypeDnaNormalized == null) subtypeDnaNormalized = Meludi.getSubtypeInfo('DNA_NORMALIZED_ALIQUOT');
*/
	
		var url = '&resetTemporary=1';
/*
		url += '&tmpfilter:INT:itemSubtype='+subtypeDna.id+'|'+subtypeDnaNormalized.id;
*/
		url += '&tmpfilter:INT:itemSubtype='+subtypeDna.id;
		url += '&tmpfilter:DATE:creationEvent.eventDate='+encodeURIComponent('<>');
		url += '&tmpfilter:FLOAT:remainingQuantity='+encodeURIComponent('>=1');
		url += '&'+encodeURIComponent('tmpfilter:STRING:&childCreationEvents(event.bioMaterial.name)')+'='+encodeURIComponent('<>%.m');

		Dialogs.selectItem('EXTRACT', 'plate', 1, url);
	}

	/**
		Callback method for manual selection.
	*/
	inspectstartplate.manuallySelected = function(event)
	{
		names[names.length] = event.detail.name;
		
		if (event.detail.remaining == 0 && names.length > 0)
		{
			Dna.loadInfoByNames(names);

			var wellsArray = [];
			var numItemsPlaced = names.length;
			if (currentSelected.length < names.length)
			{
				numItemsPlaced = currentSelected.length;
			}
			for (var i = 0; i < numItemsPlaced; i++)
			{
				var well = currentSelected[i];
				wellsArray[i] = well;
				var name = names[i];
//alert("inspect_start_plate.js::manuallySelected(): currentSelected.length = " + currentSelected.length + " names.legnth = " + names.length + " i = " + i + " name = \"" + name + "\" well = " + JSON.stringify(well));				
				// Create a new DNA object
				var dna = Dna.createByName(name);
				dna.name = name + '.fpa';
				Dna.unflag(dna);
				well.setExtract(dna);
				well.selected = false;
				// Place FPB sample in corresponding well on right side of plate
				var row = well.row;
				var column = well.column;
				well = Plate.getWell(row, column + 6);
				wellsArray[i+numItemsPlaced] = well;
				// Create a new DNA object
				dna = Dna.createByName(name);
				dna.name = name + '.fpb';
				Dna.unflag(dna);
				well.setExtract(dna);
				well.selected = false;
			}
			Plate.checkReplicates();
/*
			Plate.paint(currentSelected);
*/
			Plate.paint(wellsArray);
			inspectstartplate.updateNumFlaggedDna();
			names = [];
		}
	}

	
	/**
		Let the wizard automatically select among unprocessed DNA items.
	*/	
	inspectstartplate.autoSelect = function(event)
	{
		var frm = document.forms['meludi'];
		var wells = Plate.getWells();
		var selected = [];
		var ignore = [];
		var notEmpty = 0;
		
		// Count selected and non-empty wells and keep track of DNA that is already present
		for (var i = 0; i < wells.length; i++)
		{
			var well = wells[i];
			if (well.selected) 
			{
				selected[selected.length] = well;
				if (well.extract && !well.hasError()) notEmpty++;
			}
			if (well.extract && well.extract.id)
			{
				ignore[ignore.length] = well.extract.id;
			}
		}
		// Ignore DNA already flagged
		var flagged = Dna.getFlagged();
		for (var i = 0; i < flagged.length; i++)
		{
			var dna = flagged[i];
			if (dna.id) ignore[ignore.length] = dna.id;
		}
		
		if (selected.length == 0)
		{
			Forms.showNotification(event.currentTarget, 'Please select one or more wells were DNA should be placed.');
			return;
		}
			
		if (notEmpty > 0)
		{
			if (!confirm('Replace DNA in ' + notEmpty + ' wells?'))
			{
				return;
			}
		}
			
		var url = 'auto_select_dna.jsp?ID='+App.getSessionId();
		url += '&numToSelect='+selected.length;
/*
		url += '&quantity='+encodeURIComponent(frm.quantity_regular.value);
*/
		url += '&ignore='+ignore.join(',');
		Dialogs.openPopup(url, 'AutoSelectDNA', 750, 500);
	}

	/**
		Callback function for auto-select.
	*/	
	inspectstartplate.dnaSelected = function(response)
	{
		// Cache DNA information
		Dna.cacheInfo(response.dna);
		Dna.cacheInfo(response.flagged);
		
		inspectstartplate.updateNumFlaggedDna();
		var selected = Plate.getSelected();

/*
		// Get selected DNA names
		for (var i = 0; i < response.dna.length; i++)
		{
			if (i < selected.length)
			{
				names[i] = response.dna[i].name;
			}
		}
		if (names.length > 0)
		{
			Dna.loadInfoByNames(names);

			for (var i = 0; i < names.length && i < selected.length; i++)
			{
				var well = selected[i];
				var name = names[i];
				
				// Create a new DNA object
				var dna = Dna.createByName(name);
				Dna.unflag(dna);
				well.setExtract(dna);
				well.selected = false;
			}
			
			Plate.checkReplicates();
			Plate.paint(selected);
			inspectstartplate.updateNumFlaggedDna();
			names = [];
		}
*/

		var wellsArray = [];			
		for (var i = 0; i < response.dna.length; i++)
		{
			if (i < selected.length)
			{
				var well = selected[i];
				wellsArray[i] = well;
				// Create a new DNA object
				var name = response.dna[i].name;
				var dna = Dna.createByName(name);
				dna.name = name + '.fpa';
				//Dna.unflag(dna);
				well.setExtract(dna);
				well.selected = false;
				// Place FPB sample in corresponding well on right side of plate
				var row = well.row;
				var column = well.column;
				well = Plate.getWell(row, column + 6);
				wellsArray[i+selected.length] = well;
				// Create a new DNA object
				dna = Dna.createByName(name);
				dna.name = name + '.fpb';
				//Dna.unflag(dna);
				well.setExtract(dna);
				well.selected = false;
			}
		}
		// Check for replicates on the whole plate and repaint it
		Plate.checkReplicates(null, true);
		Plate.paint(Plate.getWells());
	}
	
	// Update the 'Flagged DNA' information
	inspectstartplate.updateNumFlaggedDna = function()
	{
		var numFlagged = Dna.getFlagged().length;
		if (numFlagged > 0)
		{
			Doc.element('flaggedDna').innerHTML = numFlagged + ' flagged DNA items.';
			Doc.show('flaggedDna');
		}
		else
		{
			Doc.hide('flaggedDna');
		}
	}

	/**
		Add 'Stratagene' to the selected wells.
	*/
	inspectstartplate.setToStratagene = function(event)
	{
		var frm = document.forms['meludi'];
		inspectstartplate.setToSpecial(event, frm.stratagene.value, true);
	}


	/**
		Add 'External DNA' to the selected wells.
	*/
	inspectstartplate.setToExternal = function(event)
	{
		inspectstartplate.setToSpecial(event, EXTERNAL_DNA_NAME, false);	
	}
	
	/**
		Add special DNA to the selected wells.
	*/
	inspectstartplate.setToSpecial = function(event, specialName, qc)
	{
		var wells = Plate.getSelected();
		
		if (wells.length == 0)
		{
			Forms.showNotification(event.currentTarget, 'No wells have been selected');
			return;
		}
		
		// Count non-empty wells
		var count = 0;
		for (var i = 0; i < wells.length; i++)
		{
			var well = wells[i];
			if (well.extract && !well.hasError()) count++;
		}
		
		if (count > 0)
		{
			if (!confirm('Replace DNA in ' + count + ' wells with '+specialName+'?'))
			{
				return;
			}
		}
		
		var info = Dna.infoByName(specialName);
		if (!info.id)
		{
			alert('Could not find any DNA with name=\''+specialName+'\'. Please check that it exists on the server.');
			return;
		}
		
		for (var i = 0; i < wells.length; i++)
		{
			var well = wells[i];
			var dna = Dna.createByInfo(info);
			dna.qc = qc;
			well.setExtract(dna);
			well.selected = false;
		}
		
		Plate.paint(wells);
		
		// Check for replicated DNA if some have been replaced with Stratagene
		if (count > 0) Plate.checkReplicates();
	}
	
	/**
		Toggle the QC flag for the selected wells. The first non-empty
		well is toggled and the rest of the wells will use the same new QC
		value.
	*/
	inspectstartplate.toggleQc = function(event)
	{
		var wells = Plate.getSelected();
	
		if (wells.length == 0) 
		{
			Forms.showNotification(event.currentTarget, 'No wells have been selected');
			return;
		}
		
		var gotQc = false;
		var newQc;
		var count = 0;
		for (var i = 0; i < wells.length; i++)
		{
			var well = wells[i];
			if (well.extract)
			{
				// Toggle QC flag for the first well with DNA, then use the same flag for the rest
				if (!gotQc)
				{
					gotQc = true;
					newQc = !well.extract.qc;
				}
				well.extract.qc = newQc;
				count++;
			}
		}
		
		if (count == 0)
		{
			Forms.showNotification(event.currentTarget, 'None of the selected wells contain any DNA');
		}
		else
		{
			Plate.paint(wells);
		}
	}

	// Set a comment on the selected wells
	inspectstartplate.commentSelected = function(event)
	{
		var wells = Plate.getSelected();
		
		if (wells.length == 0)
		{
			Forms.showNotification(event.currentTarget, 'No wells have been selected');
			return;
		}
		
		var count = 0;
		var comment = '';
		for (var i = 0; i < wells.length; i++)
		{
			var well = wells[i];
			if (well.extract) 
			{
				count++;
				if (well.extract.comment) comment = well.extract.comment;
			}
		}
			
		if (count == 0)
		{
			Forms.showNotification(event.currentTarget, 'None of the selected wells contain any DNA');
			return;
		}

		comment = prompt('Comment', comment || '');
		if (comment == null) return;
		
		if (comment == '') comment = null;
		for (var i = 0; i < wells.length; i++)
		{
			var well = wells[i];
			if (well.extract) well.extract.comment = comment;
		}
		Plate.paint(wells);
	}

	/**
		Empty the selected wells from DNA. They can later be pasted again.
	*/
	inspectstartplate.cutSelected = function(event)
	{
		var wells = Plate.getSelected();
		
		if (wells.length == 0)
		{
			Forms.showNotification(event.currentTarget, 'No wells have been selected');
			return;
		}
		
		var count = 0;
		var valid = 0;
		for (var i = 0; i < wells.length; i++)
		{
			var well = wells[i];
			if (well.extract) 
			{
				count++;
				if (!well.hasError()) valid++;
			}
		}
		
		if (count == 0)
		{
			Forms.showNotification(event.currentTarget, 'None of the selected wells contain any DNA');
			return;
		}
		
		// Ask for confirmation before deleting from valid wells
		if (valid > 0)
		{
			if (!confirm('Clear DNA from ' + count + ' of ' + wells.length + ' selected wells?'))
			{
				return;
			}
		}
	
		inspectstartplate.copySelected();
/*
		for (var i = 0; i < wells.length; i++)
		{
			var well = wells[i];
			well.setExtract(null);
			well.selected = false;
		}
		Plate.paint(wells);
*/
		var wellsArray = [];			
		for (var i = 0; i < wells.length; i++)
		{
			var well = wells[i];
			well.setExtract(null);
			well.selected = false;
			wellsArray[i] = well;
			// Clear FPB sample in corresponding well on right side of plate
			var row = well.row;
			var column = well.column;
			well = Plate.getWell(row, column + 6);
			well.setExtract(null);
			well.selected = false;
			wellsArray[i+wells.length] = well;
		}
		Plate.paint(wellsArray);
		Plate.checkReplicates();
	}
	
	var copy;
	/**
		Copy information about the selected wells. 
	*/
	inspectstartplate.copySelected = function(event)
	{
		// Clear existing wells in the copy
		var repaint = [];
		if (copy && copy.length > 0)
		{
			for (var i = 0; i < copy.length; i++)
			{
				var cp = copy[i];
				cp.well.doneWithCopy();
				repaint[repaint.length] = cp.well;
			}
		}
		Plate.paint(repaint);
		
		// Place selected wells in the copy
		var wells = Plate.getSelected();
		copy = [];
		for (var i = 0; i < wells.length; i++)
		{
			var well = wells[i];
			copy[copy.length] = well.makeCopy();
			well.selected = false;
		}
		Plate.paint(wells);
	}
	
	/**
		Paste information into the selected wells.
	*/
	inspectstartplate.pasteToSelected = function(event)
	{
		if (!copy || copy.length == 0) 
		{
			Forms.showNotification(event.currentTarget, 'Nothing to paste. Please cut or copy wells first.');
			return;
		}
		
		var wells = Plate.getSelected();
		// Count non-empty and valid wells
		var count = 0;
		for (var i = 0; i < wells.length; i++)
		{
			var well = wells[i];
			if (well.extract && !well.hasError()) count++;
		}
		
		if (count > 0)
		{
			if (!confirm('Replace DNA in ' + count + ' wells with copy?'))
			{
				return;
			}
		}
		
		var wi = 0;
		var ci = 0;
		var copyEmpty;
		var askIfEmpty = true;
		var repaint = [];
		
		while (wi < wells.length && ci < copy.length)
		{
			var well = wells[wi];
			var cp = copy[ci];
			var dna = null;
			var name = null;
			cp.well.doneWithCopy();
			repaint[repaint.length] = cp.well;
			if (cp.name)
			{
				name = cp.name;
				var suffixIndex = name.indexOf('.fpa');
				if (suffixIndex >= 0)
				{
					name = name.substring(0, suffixIndex);
				}
/*
				dna = Dna.createByName(cp.name);
*/
				dna = Dna.createByName(name);
				dna.comment = cp.comment;
			}
			else
			{
				// The copy is from an empty well
				if (askIfEmpty)
				{
					askIfEmpty = false;
					copyEmpty = confirm('Do you want to copy empty wells? If not, only non-empty well are copied.');
				}
			}
			if (dna != null || copyEmpty)
			{
				// FPA
				if (dna != null)
				{
					dna.name += '.fpa'; 
				}
				well.setExtract(dna);
				well.selected = false;
				repaint[repaint.length] = well;
				wi++;
				// FPB sample in corresponding well on right side of plate
				var row = well.row;
				var column = well.column;
				well = Plate.getWell(row, column + 6);
				dna = null;
				if (name)
				{
					dna = Dna.createByName(name);
					dna.comment = cp.comment;
				}
				if (dna != null)
				{
					dna.name += '.fpb'; 
				}
				well.setExtract(dna);
				well.selected = false;
				repaint[repaint.length] = well;
			}
			ci++;
		}
		Plate.paint(repaint);
		copy.splice(0, ci);
		Plate.checkReplicates();
	}

	/*
		If exactly two wells have been selected, switch the DNA in them.
	*/
	inspectstartplate.switchSelected = function(event)
	{
		var wells = Plate.getSelected();
	
		if (wells.length != 2)
		{
			Forms.showNotification(event.currentTarget, 'Exactly 2 wells must be selected.');
			return;
		}

/*	
		var dna0 = wells[0].extract;
		var dna1 = wells[1].extract;
		
		wells[0].setExtract(dna1);
		wells[1].setExtract(dna0);
		Plate.paint(wells);
*/
		var wellsArray = [];
		// FPA
		wellsArray[0] = wells[0];
		wellsArray[1] = wells[1];

		var dna0 = wells[0].extract;
		var dna1 = wells[1].extract;
		
		wells[0].setExtract(dna1);
		wells[1].setExtract(dna0);
		// FPB samples in corresponding wells on right side of plate
		var wellsFpb = [];
		for (var i=0; i<2; i++)
		{	
			var row = wells[i].row;
			var column = wells[i].column;
			wellsFpb[i] = Plate.getWell(row, column + 6);
		}
		wellsArray[2] = wellsFpb[0];
		wellsArray[3] = wellsFpb[1];

		dna0 = wellsFpb[0].extract;
		dna1 = wellsFpb[1].extract;
		
		wellsFpb[0].setExtract(dna1);
		wellsFpb[1].setExtract(dna0);
		//
		Plate.paint(wellsArray);
	}

	// Toggle the 'special select' menu on and off
	inspectstartplate.toggleSpecialSelect = function(event)
	{
		var pos = Doc.getElementPosition('iconSpecialSelect');
		Menu.toggleTopMenu('menuSpecialSelect', pos.left+pos.width/4, pos.top+pos.height); 
		event.stopPropagation();
	}
	
	/**
		Open a pop-up dialog for manual selection of extracts.
	*/
	inspectstartplate.manualSelectExtract = function(event)
	{
		if (subtypeDna == null) subtypeDna = Meludi.getSubtypeInfo('DNA');
	
		var url = '&resetTemporary=1';
		url += '&tmpfilter:INT:itemSubtype='+subtypeDna.id;
		url += '&tmpfilter:STRING:name='+encodeURIComponent('%.d');
		url += '&tmpfilter:DATE:creationEvent.eventDate='+encodeURIComponent('');
		url += '&tmpfilter:FLOAT:remainingQuantity='+encodeURIComponent('>0');
		url += '&'+encodeURIComponent('tmpfilter:STRING:&childCreationEvents(event.bioMaterial.name)')+'='+encodeURIComponent('');

		Dialogs.selectItem('EXTRACT', 'plate', 1, url);
	}

	/**
		Open a pop-up dialog for manual selection of controls.
	*/
	inspectstartplate.manualSelectControl = function(event)
	{
		if (subtypeDna == null) subtypeDna = Meludi.getSubtypeInfo('DNA');
	
		var url = '&resetTemporary=1';
		url += '&tmpfilter:INT:itemSubtype='+subtypeDna.id;
		url += '&tmpfilter:STRING:name='+encodeURIComponent('Horizon%.d');
		url += '&tmpfilter:DATE:creationEvent.eventDate='+encodeURIComponent('');
		url += '&tmpfilter:FLOAT:remainingQuantity='+encodeURIComponent('>0');
		url += '&'+encodeURIComponent('tmpfilter:STRING:&childCreationEvents(event.bioMaterial.name)')+'='+encodeURIComponent('');

		Dialogs.selectItem('EXTRACT', 'plate', 1, url);
	}

	var contextDNA = null;
	var contextX;
	var contextY;
	/**
		Reacts to 'mouseup' and 'contextmenu' events for the bioplate. 
		This should bring up the cut/copy/paste context menu depending on which
		mouse button that was clicked.
	*/
	inspectstartplate.contextEvent = function(event)
	{
		var frm = document.forms['meludi'];
		// Context menu on 'right' mouse button
		// Can't just check the button since two events are sent ('mouseup' and 'contextmenu')
		var showContext = event.type == 'contextmenu' && event.button == 2;
			
		if (showContext)
		{
			event.preventDefault(); // Prevents the default right-click menu from appearing
			
			// Get the well that is right-clicked and the DNA that is in it
			contextDNA = null;
			var well = event.target;
			while (well && (!well.id || well.id.indexOf('well') != 0))
			{
				well = well.parentNode;
			}
			if (well)
			{
				var c = well.id.split(/\./);
				contextDNA = Plate.getWell(parseInt(c[1]), parseInt(c[2])).extract;
			}
			// Update the context meny
			var caseSummaryMenu = Doc.element('mnuCaseSummary');
			if (contextDNA)
			{
				caseSummaryMenu.title = 'Show case summary for ' + Strings.encodeTags(contextDNA.name);
				Doc.show('sepCaseSummary');
				Doc.show('mnuCaseSummary');
			}
			else
			{
				Doc.hide('sepCaseSummary');
				Doc.hide('mnuCaseSummary');
			}
			
			var menu = Doc.element('menuContext');
			// 1 pixel offset to avoid losing well focus outline
			contextX = event.clientX+1;
			contextY = event.clientY+1;
			// Need short delay since 'mouseup' are also sent as 'click' events
			// to the 'document' object which BASE already have a Menu.hideAll()
			// call which would hide the menu immediately
			setTimeout(inspectstartplate.showContextMenu, 100);
		}
	}
	
	inspectstartplate.showContextMenu = function()
	{
		Menu.showTopMenu('menuContext', contextX, contextY);
	}

	inspectstartplate.showCaseSummary = function()
	{
		if (!contextDNA) return;
		var caseName = contextDNA.name;
		// Remove suffix starting with dot "."
		var firstDotIndex = caseName.indexOf(".");
		if (firstDotIndex >= 0)
		{
			caseName = caseName.substring(0, firstDotIndex);
		}
		var url = '../reports/case_summary.jsp?ID='+App.getSessionId();
		url += '&caseName='+encodeURIComponent(caseName);
		url += '&pageType=popup';
		Dialogs.openPopup(url, 'CaseSummary'+caseName, 1000, 700);
	}
	
	inspectstartplate.showFlaggedDna = function()
	{
		currentSelected = Plate.getSelected();
		currentIndex = 0;

		var flagged = Dna.getFlagged();
		var url = 'show_flagged_dna.jsp?ID='+App.getSessionId();
		url += '&numFlagged='+flagged.length;
		url += '&numSelected='+currentSelected.length;
		Dialogs.openPopup(url, 'FlaggedDna', 800, 500);
	}

	
	//Flag the selected DNA
	inspectstartplate.flagSelected = function(event)
	{
		var flag = Data.get(event.currentTarget, 'flag');
		var wells = Plate.getSelected();
		
		if (wells.length == 0)
		{
			Forms.showNotification(event.currentTarget, 'No wells have been selected.');
			return;
		}
		
		var comment = null;
		if (flag == 'ManualFlag')
		{
			// Check if a comment has been set on any of the DNA
			for (var i = 0; i < wells.length; i++)
			{
				var well = wells[i];
				if (well.extract && well.extract.info) 
				{
					if (well.extract.info.comment) comment = well.extract.info.comment;
				}
			}
			comment = prompt('Comment', comment || '');
			if (!comment) return;
		}
		
		var count = 0;
		for (var i = 0; i < wells.length; i++)
		{
			var well = wells[i];
			var dna = well.extract;
			if (dna && !dna.stratagene && !dna.external)
			{
				Dna.flag(dna, flag);
				dna.info.comment = comment;
				well.setExtract(null);
				count++;
			}
			well.selected = false;
		}
		
		inspectstartplate.updateNumFlaggedDna();
		Plate.paint(wells);
	}

	inspectstartplate.downloadLibPrepFile = function()
	{
		var frm = document.forms['meludi'];
		var isNeoPrep = Data.int('page-data', 'is-neoprep');

		var submitInfo = {};
		
		var libraryPreparationProtocolName = '';
		var libraryPreparationProtocol = frm.libraryPreparationProtocol.value;
		if (libraryPreparationProtocol)
		{
			libraryPreparationProtocolName = protocolMap[libraryPreparationProtocol];
		}
		submitInfo.libraryPreparationProtocol = parseInt(libraryPreparationProtocol, 10);
		submitInfo.libraryPreparationProtocolName = libraryPreparationProtocolName;
		submitInfo.bioplate = inspectstartplate.fetchBioplateData();

		var url = '../LibPrep.servlet?ID='+App.getSessionId();
		url += '&cmd=PrepareDownloadLibPrepFile';
		url += '&referenceName='+encodeURIComponent(frm.plateName.value);

		// POST
		Wizard.showLoadingAnimation('Performing registration...');
		Wizard.asyncJsonRequest(url, inspectstartplate.downloadLibPrepFileResults, 'POST', JSON.stringify(submitInfo));
	}

	inspectstartplate.fetchBioplateData = function()
	{
		var frm = document.forms['meludi'];
		var isNeoPrep = Data.int('page-data', 'is-neoprep');
		var schema = PoolSchema.getById(frm.pool_schema.value);

		var plateInfo = {};
		plateInfo.id = Plate.id;
		plateInfo.name = Plate.name;
		plateInfo.comments = frm.comments.value;
		plateInfo.kitName = frm.tsLibPrepKit.value;
		plateInfo.kitId = frm.hiddenLibPrepKitId.value;
		plateInfo.plateType = isNeoPrep ? 'NEOPREP' : 'DNA';
		plateInfo.poolSchema = schema ? schema.id : null;
		if (isNeoPrep)
		{
			plateInfo.barcodeVariant = frm.barcode_variant.value;
		}
		plateInfo.wells = [];

		var wells = Plate.getWells();
		for (var i = 0; i < wells.length; i++)
		{
			var well = wells[i];
			var dna = well.extract;
			if (dna && dna.info.origId)
			{
				var tmp = {};
				tmp.row = well.row;
				tmp.column = well.column;
				tmp.dna = {};
				tmp.dna.id = dna.info.origId;
				tmp.dna.usedQuantity = dna.usedQuantity;
				tmp.dna.dilutionConc = dna.dilutionConc;
				tmp.dna.qc = dna.qc;
				tmp.dna.comment = dna.comment;
				if (well.barcode)
				{
					tmp.dna.barcode = well.barcode;
				}
				plateInfo.wells[plateInfo.wells.length] = tmp;
			}
		}
		
		return plateInfo;
	}

	inspectstartplate.downloadLibPrepFileResults = function(response)
	{
		var tmpFilePath = response;
		var frm = document.forms['meludi'];
		var url = '../LibPrep.servlet?ID='+App.getSessionId();
		url += '&cmd=DownloadLibPrepFile';
		url += '&referenceName='+encodeURIComponent(frm.plateName.value);
		url += '&tmpFilePath='+encodeURIComponent(tmpFilePath);
		window.open(url);
	}

	inspectstartplate.viewProtocol = function(event)
	{
		var frm = document.forms['meludi2'];
		var bioplateData = inspectstartplate.fetchBioplateData();
		var submitInfo = inspectstartplate.fetchSubmitInfo();

		// Save values in hidden input fields for retrieval after HTML request
		var submitInfoJsonStr = JSON.stringify(submitInfo, inspectstartplate.circular_reference_remover);
		// Enable garbage collection
		circular_reference_cache = null;
		frm.hiddenSubmitInfo.value = submitInfoJsonStr;
		frm.hiddenNumItems.value = submitInfo.items.length;
		frm.view.value = Data.get(event.currentTarget, 'protocol-type');
		frm.bioplateName.value = bioplateData.name;
		frm.bioplateDescription.value = bioplateData.comments;
		frm.bioplateKitName.value = bioplateData.kitName;

		// Calling frm.submit() will change jsp file to libprep_plate_protocol2.jsp
		frm.submit();
	}

	inspectstartplate.createProtocol = function()
	{
		var frm = document.forms['meludi'];
		var submitInfo = inspectstartplate.fetchSubmitInfo();

		// Save values in hidden input fields for retrieval after HTML request
		var submitInfoJsonStr = JSON.stringify(submitInfo, inspectstartplate.circular_reference_remover);
		// Enable garbage collection
		circular_reference_cache = null;
		frm.hiddenSubmitInfo.value = submitInfoJsonStr;
		frm.hiddenNumItems.value = submitInfo.items.length;

		// Calling frm.submit() will change jsp file to libprep_dilution_protocol2.jsp
		frm.submit();
	}

	inspectstartplate.fetchSubmitInfo = function()
	{
		var frm = document.forms['meludi'];
		var submitInfo = {};
		submitInfo.items = [];

		submitInfo.docMode = 'report';
		// Get list of selected extract source items
		var selItemsList = [];
		for (var column = 0; column < 6; column++)
		{
			for (var row = 0; row < 8; row++)
			{
				var well = Plate.getWell(row, column);
				var dna = well.extract;
				if (dna != null)
				{				
					dna.bioWell = well;
					dna.fpaPos = inspectstartplate.wellRowColumnToWellStr(row, column); 
					dna.fpbPos = inspectstartplate.wellRowColumnToWellStr(row, column + 6); 
					selItemsList[selItemsList.length] = dna;
				}
			}
		}
		var numItems = selItemsList.length;
		var totNumItems = numItems;
		for (var i = 0; i < totNumItems; i++)
		{
			var item = null;
			item = selItemsList[i];
			submitInfo.items[submitInfo.items.length] = item;
		}

		// Library preparation dilution
		var libraryPreparationProtocolName = '';
		var libraryPreparationProtocol = frm.libraryPreparationProtocol.value;
		if (libraryPreparationProtocol)
		{
			libraryPreparationProtocolName = protocolMap[libraryPreparationProtocol];
		}
		submitInfo.libraryPreparationProtocol = parseInt(libraryPreparationProtocol, 10);
		submitInfo.libraryPreparationProtocolName = libraryPreparationProtocolName;
		submitInfo.startPlateName = frm.plateName.value;
		var now = new Date();
		var dilutionProtocolDate = inspectstartplate.formatDate(now);
		submitInfo.dilutionProtocolDate = dilutionProtocolDate;

		return submitInfo;
	}
	
var circular_reference_cache = [];

	inspectstartplate.circular_reference_remover = function(key, value)
	{
		if (typeof value === 'object' && value !== null)
		{
			if (circular_reference_cache == null)
			{
				circular_reference_cache = [];
			}
			if (circular_reference_cache.indexOf(value) !== -1)
			{
				// Circular reference found, discard key
				return;
			}
			// Store value in collection
			circular_reference_cache.push(value);
		}
		return value;
	}

	inspectstartplate.initElements = function(element, autoInit)
	{
		if (autoInit == 'plate-col')
		{
			Events.addEventHandler(element, 'click', inspectstartplate.toggleColumn);
			Events.addEventHandler(element, 'mouseover', inspectstartplate.highlightColumn);
			Events.addEventHandler(element, 'mouseout', inspectstartplate.highlightColumn);
		}
		else if (autoInit == 'plate-row')
		{
			Events.addEventHandler(element, 'click', inspectstartplate.toggleRow);
			Events.addEventHandler(element, 'mouseover', inspectstartplate.highlightRow);
			Events.addEventHandler(element, 'mouseout', inspectstartplate.highlightRow);
		}
		else if (autoInit == 'plate-well')
		{
			Events.addEventHandler(element, 'click', inspectstartplate.toggleWell);
			Events.addEventHandler(element, 'mouseover', inspectstartplate.highlightReplicated);
			Events.addEventHandler(element, 'mouseout', inspectstartplate.highlightReplicated);
		}
		else if (autoInit == 'special-select')
		{
			Events.addEventHandler(element, 'click', inspectstartplate.specialToggle)
		}
	}

	// Restrict wells to left side of plate
	inspectstartplate.restrictWells = function(wellsInput)
	{
		var wells = [];
		for (var i = 0; i < wellsInput.length; i++)
		{
			well = wellsInput[i];
			if (well.column < 6)
			{
				wells[wells.length] = well;
			}
		}
		return wells;
	}

	// Toggle the selected status of a single well
	inspectstartplate.toggleWell = function(event)
	{
		var row = Data.int(event.currentTarget, 'row');
		var column = Data.int(event.currentTarget, 'col');
		if (column >= 6)
		{
			return;
		}
		var well = Plate.getWell(row, column);
		Plate.toggleSelected([well]);
	}

	// Toggle the selected status of a complete row
	inspectstartplate.toggleRow = function(event)
	{
		var row = Data.int(event.currentTarget, 'row');
/*
		Plate.toggleSelected(Plate.getRow(row));
*/
		var wells = inspectstartplate.restrictWells(Plate.getRow(row));
		Plate.toggleSelected(wells);
	}

	// Toggle the selected status of a complete column
	inspectstartplate.toggleColumn = function(event)
	{
		var column = Data.int(event.currentTarget, 'col');
		if (column >= 6)
		{
			return;
		}
		Plate.toggleSelected(Plate.getColumn(column));
	}
	
	// Toggle the selected status of a pool
	inspectstartplate.togglePool = function(event)
	{
		var pool = Data.int(event.target, 'pool-num');
/*
		if (!isNaN(pool)) Plate.toggleSelected(Plate.getPool(pool));
*/
		if (!isNaN(pool) && pool < 1)
		{
			Plate.toggleSelected(Plate.getPool(pool));
		}
	}
	
	// Highlight enable/disable all wells in a column
	inspectstartplate.highlightColumn = function(event)
	{
		var column = Data.int(event.currentTarget, 'col');
		if (column >= 6)
		{
			return;
		}
		var on = event.type == 'mouseover';
		
		Doc.addOrRemoveClass(event.currentTarget, 'highlight-column', on);
		var wells = Plate.getColumn(column);
		Plate.setHighlight(wells, 'highlight-column', on);
	}

	// Highlight enable/disable all wells in a row
	inspectstartplate.highlightRow = function(event)
	{
		var row = Data.int(event.currentTarget, 'row');
		var on = event.type == 'mouseover';
		
		Doc.addOrRemoveClass(event.currentTarget, 'highlight-row', on);
/*
		var wells = Plate.getRow(row);
*/
		var wells = inspectstartplate.restrictWells(Plate.getRow(row));
		Plate.setHighlight(wells, 'highlight-row', on);
	}
	
	// Highligt enable/disable all wells in a pool
	inspectstartplate.highlightPool = function(event)
	{
		var pool = Data.int(event.target, 'pool-num');
		var on = event.type == 'mouseover';
/*
		if (!isNaN(pool))
*/
		if (!isNaN(pool) && pool < 1)
		{
			Doc.addOrRemoveClass(event.target, 'highlight-pool', on);
			var wells = Plate.getPool(pool);
			Plate.setHighlight(wells, 'highlight-pool', on);
		}
	}

	/**
		Highlight all replicated wells with the same DNA as the given well.
	*/
	inspectstartplate.highlightReplicated = function(event)
	{
		var column = Data.int(event.currentTarget, 'col');
		if (column >= 6)
		{
			return;
		}
		var row = Data.int(event.currentTarget, 'row');
		var on = event.type == 'mouseover';
		
		var well = Plate.getWell(row, column);
		if (well.extract && well.replicate)
		{
			// Get center coordinates for the current well
			var pos = Doc.getElementPosition(well.tag);
			var jsPos = new jsPoint(pos.left+pos.width/2, pos.top+pos.height/2);
	
			var replicated = Plate.getWellsByName(well.extract.name);
			for (var i = 0; i < replicated.length; i++)
			{
				var rep = replicated[i];
				if (rep != well)
				{
					Doc.addOrRemoveClass(rep.tag, 'highlight-replicated', on);
					if (rep.line)
					{
						// Clear any recent lines
						graphics.clearDrawing(rep.line);
						rep.line = null;
					}
					if (on)
					{
						// We draw a line between the current and replicated well
						var rPos = Doc.getElementPosition(rep.tag);
						rep.line = graphics.drawLine(pen, jsPos, new jsPoint(rPos.left+rPos.width/2, rPos.top+rPos.height/2));
					}
				}
			}
		}
	}

	// Some special toogle operations
	inspectstartplate.specialToggle = function(event)
	{
		var what = Data.get(event.currentTarget, 'special');
		var wells = [];
		if (what == 'all' || what == 'empty' || what == 'none')
		{
			// All wells or all empty (will be filtered later)
			wells = Plate.getWells();
		}
		else if (what == 'pools' || what == 'empty-pools')
		{
			// All primary pools or all empty in the primary pools (will be filtered later)
			for (var i = 0; i < Plate.poolSchema.numPools; i++)
			{
				wells = wells.concat(Plate.getPool(i));
			}
		}
		else if (what == 'stratagene')
		{
			// All wells with 'Stratagene'
			var tmp = Plate.getWells();
			for (var i = 0; i < tmp.length; i++)
			{
				var well = tmp[i];
				if (well.extract && well.extract.stratagene) wells[wells.length] = well;
			}
		}
		else if (what == 'external')
		{
			// All wells with 'External DNA'
			var tmp = Plate.getWells();
			for (var i = 0; i < tmp.length; i++)
			{
				var well = tmp[i];
				if (well.extract && well.extract.external) wells[wells.length] = well;
			}
		}
		else if (what == 'replicates')
		{
			// All wells with replicated DNA
			var tmp = Plate.getWells();
			for (var i = 0; i < tmp.length; i++)
			{
				var well = tmp[i];
				if (well.extract && well.replicate) wells[wells.length] = well;
			}
		}
		else if (what == 'error')
		{
			// All wells with an error
			var tmp = Plate.getWells();
			for (var i = 0; i < tmp.length; i++)
			{
				var well = tmp[i];
				if (well.hasError()) wells[wells.length] = well;
			}
		}
		else if (what == 'warning')
		{
			// All wells with a warning
			var tmp = Plate.getWells();
			for (var i = 0; i < tmp.length; i++)
			{
				var well = tmp[i];
				if (well.warning) wells[wells.length] = well;
			}
		}
		
		// Extra filter for empty wells only
		if (what.indexOf('empty') != -1)
		{
			var tmp = wells;
			wells = [];
			for (var i = 0; i < tmp.length; i++)
			{
				if (!tmp[i].extract) wells[wells.length] = tmp[i];
			}
		}
		
		if (what == 'none')
		{
			Plate.setSelected(wells, false);
		}
		else
		{
			Plate.toggleSelected(wells);
		}
	}

	// Format value as a date in format yyyy-mm-dd
	inspectstartplate.asDate = function(value)
	{
		if (!value) return '';
		if (value.length == 8)
		{
			value = value.substr(0, 4) + '-' + value.substr(4, 2) + '-' + value.substr(6, 2);
		}
		return value;
	}

	// Format date value in format yyyy-mm-dd
	inspectstartplate.formatDate = function(date)
	{
		if (!date) return '';
		var year = date.getFullYear();
		var month = (date.getMonth() + 1);
		var day = date.getDate();
		var monthStr = '' + month;
		if (month < 10)
		{
			monthStr = '0' + month;
		}
		var dayStr = '' + day;
		if (day < 10)
		{
			dayStr = '0' + day;
		}
		var dateStr = '' + year + '-' + monthStr + '-' + dayStr;
		return dateStr;
	}

	return inspectstartplate;
}();

Doc.onLoad(InspectStartPlate.initPage);
Doc.addElementInitializer(InspectStartPlate.initElements);


var Dna = function()
{
	var dna = {};
	var flagged = [];
	var info = [];
	
	/**
		Create a new DNA object by name. More information
		about the DNA is automatically loaded from the database.
	*/
	dna.createByName = function(name)
	{
		var tmp = {};
		tmp.name = name;
		tmp.stratagene = Meludi.isStratagene(name)
		tmp.external = Meludi.isExternal(name);
		tmp.info = dna.infoByName(name);
//alert("inspect_start_plate.js::dna.createByName(name): name = " + name + " info = " + JSON.stringify(tmp.info));
		tmp.id = tmp.info.id;
		return tmp;
	}
	
	/**
		Create a new DNA object by parent name. More information
		about the DNA is automatically loaded from the database.
	*/
	dna.createByParentName = function(parentName, name, id)
	{
//alert("inspect_start_plate.js::dna.createByParentName(parentName, name, id): parentName = " + parentName + " name = " + name + " id = " + id);
		var tmp = {};
		tmp.name = name;
		tmp.id = id;
		tmp.stratagene = Meludi.isStratagene(parentName)
		tmp.external = Meludi.isExternal(parentName);
		var parentDna = dna.infoByName(parentName);
		// Copy some info to original info
		var info = {};
		info.origBioWell = parentDna.bioWell;
		info.origQubitConc = parentDna.QubitConc;
		info.origDeltaCt = parentDna.deltaCt;
		info.origId = parentDna.id;
		info.origName = parentDna.name;
		info.origDescription = parentDna.comment;
		// Reset info for created item
		info.bioWell = null;
		info.QubitConc = null;
		info.deltaCt = null;
		info.id = null;
		// Set tmp.info to new JSONObject info
		tmp.info = info;
		return tmp;
	}
	
	/**
		Create a new DNA object by info object.
	*/
	dna.createByInfo = function(info)
	{
		var tmp = {};
		tmp.name = info.name;
		tmp.stratagene = Meludi.isStratagene(tmp.name);
		tmp.external = Meludi.isExternal(tmp.name);
		tmp.id = info.id;
		tmp.info = info;
		return tmp;
	}
	
	/**
		Get information about a DNA item with a given name.
	*/
	dna.infoByName = function(name)
	{
		var key = 'N'+name;
		if (!info[key])
		{
			dna.loadInfoByNames([name]);
			if (!info[key]) info[key] = {};
		}
		return info[key];
	}

	/**
		Load and cache DNA information for all DNA items with a name in the
		given list.
	*/
	dna.loadInfoByNames = function(names)
	{
		var newNames = [];
		for (var i = 0; i < names.length; i++)
		{
			if (!info['N'+names[i]]) newNames[newNames.length] = names[i];
		}
		
		var submitInfo = {};
		submitInfo.names = newNames;
		
		if (newNames.length > 0)
		{
			var url = '../Dna.servlet?ID='+App.getSessionId();
			url += '&cmd=GetDnaInfoFromNames';	
			var response = Wizard.syncJsonRequest(url, 'POST', JSON.stringify(submitInfo));
			dna.cacheInfo(response.dna);
		}
	}
	
	dna.cacheInfo = function(dnaList)
	{
		for (var i = 0; i < dnaList.length; i++)
		{
			var r = dnaList[i];
			info['N'+r.name] = r;
			info['I'+r.id] = r;
			if (r.flag) flagged[flagged.length] = r;
		}
	}

	dna.unflag = function(r)
	{
		if (!r.info || !r.info.flag) return false;

		r.info.flag = null;
		for (var i = 0; i < flagged.length; i++)
		{
			if (r.id == flagged[i].id)
			{
				flagged.splice(i, 1);
				break;
			}
		}
		return true;
	}
	
	dna.flag = function(r, flag)
	{
		if (!r.info || r.info.flag) return false;
		r.info.flag = flag;
		flagged[flagged.length] = r.info;
	}
	
	dna.getFlagged = function()
	{
		return flagged;
	}
	
	return dna;
}();



var WellPainter = function()
{
	var painter = {};
	
	// Add class indicators for replicates and QC assigned wells
	painter.getClassNameForWell = function(well)
	{
		var cls = '';
		if (well.extract)
		{
			if (well.replicate) cls += ' replicate';
			if (well.extract.qc) cls += ' qc';
		}
		if (well.column == Plate.columns-1) cls += ' last-child';
		
		return cls;
	}
	
	painter.getWellText = function(well)
	{
		var frm = document.forms['meludi'];
		var text = '';
		if (well.duplicates)
		{
			well.setError('Duplicate DNA in this location');
			text += '<div class="name">'+Strings.encodeTags(well.duplicates.join(', '))+'</div>';
		}
		else if (well.extract)
		{
			// The well contains DNA
			var dna = well.extract;
			var info = dna.info;
			// --- DNA aliquot name
			text += '<div class="name">'+Strings.encodeTags(dna.name)+'</div>';
//alert("inspect_start_plate.js::getWellText(): info = " + JSON.stringify(info));
			
			var warningMsg = [];
			// --- Bioplate location
			if (info.origBioWell)
			{
				var origDnaWell = info.origBioWell;
				text += '<div class="location">'+Strings.encodeTags(origDnaWell.bioPlate.name+'['+origDnaWell.location)+']</div>';
			}
			else if (info.preNormalized)
			{
				text += '<div class="location">PreNormalized</div>';
			}
			else if (!dna.stratagene && !dna.external)
			{
				if (!painter.hasClass('plate', 'hide-location'))
				{
					warningMsg[warningMsg.length] = 'No location';
				}
			}

			if (info && info.origId && !dna.stratagene && !dna.external)
			{
				// Calculations for each aliquot
				var origDeltaCt = info.origDeltaCt;
				var dilutionFactor = painter.fetchDilutionFactor(origDeltaCt);
				var VOL_ALIQUOT = 10.0; // µl
				var volDNA = VOL_ALIQUOT/dilutionFactor;
				var water = VOL_ALIQUOT - volDNA;
				// Prepare for future modification of solution volume depending on other parameters
				var volAliquotSolution = VOL_ALIQUOT;
				var volDnaInSolution = volAliquotSolution/dilutionFactor;
				var waterInSolution = volAliquotSolution - volDnaInSolution;
				var removedVolume = 2 * volDnaInSolution; // µl
				var origConc = info.origQubitConc; // ng/µl
				var calcConc = origConc/dilutionFactor;
				//var usedQuantityPerAliquout = Math.ceil(VOL_ALIQUOT*origConc/dilutionFactor/1000); // µg
				//var usedQuantityPerAliquout = VOL_ALIQUOT*origConc/dilutionFactor/1000; // µg
				//var usedQuantity = volDNA*origConc; // ng
				var usedQuantity = volDnaInSolution*origConc; // ng
				dna.usedQuantity = dna.qc ? QUANTITY_QC : usedQuantity/1000; // µg
//alert("inspect_start_plate.js::painter.getWellText(): name = " + dna.name + " removedVolume = " + removedVolume + " origDeltaCt = " + origDeltaCt + " dilutionFactor = " + dilutionFactor + " origConc = " + origConc + " usedQuantity = " + usedQuantity + " ng");

/*				
				dna.usedQuantity = dna.qc ? QUANTITY_QC : QUANTITY_REGULAR; // µg
				dna.dilutionConc = 1000 * dna.usedQuantity / TOTAL_VOLUME; // ng/µl

				var volDNA = Math.ceil(10000*dna.usedQuantity/info.origQubitConc) / 10; // µl, rounded to 1 decimal
				var water = TOTAL_VOLUME - volDNA;
*/				

				var MINIMAL_DNA_VOLUME = frm.min_vol_dna.value;
				if (volDnaInSolution < MINIMAL_DNA_VOLUME)
				{
/*
					// Large mix since we do not want to take less than 1µl
					volDNA = MINIMAL_DNA_VOLUME;
					dna.usedQuantity = volDNA * info.origQubitConc / 1000; // µg
					var totalVolume = 1000 * dna.usedQuantity / dna.dilutionConc; // µl
					water = totalVolume - volDNA;
					warningMsg[warningMsg.length] = 'Large mix';
*/
					warningMsg[warningMsg.length] = 'DNA vol ' + Numbers.formatNumber(volDnaInSolution, 2) + 'µl < ' + MINIMAL_DNA_VOLUME + 'µl';
				}

				// Store dilution data
				info.dilutionFactor = dilutionFactor;
				info.volAliquotSolution = volAliquotSolution;
				info.volDnaInSolution = volDnaInSolution;
				info.waterInSolution = waterInSolution;
/*				
				if (info.remainingQuantity != null)
				{
					//text += '<div class="quantity">'+Numbers.formatNumber(info.remainingQuantity, 2) + 'µg</div>';
					var quantityLeft = 1000.0*info.remainingQuantity - 2*usedQuantity; // ng
//alert("select_dna.js::painter.getWellText(): name = " + dna.name + " info.remainingQuantity = " + info.remainingQuantity + " µg usedQuantity = " + usedQuantity + " ng quantityLeft = " + quantityLeft + " ng");
					//text += '<div class="quantity">'+Numbers.formatNumber(quantityLeft, 2) + 'µg '+Numbers.formatNumber(volDNA, 1)+'µl</div>';
					text += '<div class="quantity">'+Numbers.formatNumber(usedQuantity, 2) + 'ng '+Numbers.formatNumber(volDNA, 1)+'µl</div>';

					// Must have at least 1.1µg or 1.22µg (if default values are used)
					var remainLimit = dna.qc ? QUANTITY_QC : QUANTITY_REGULAR;
*/
/*
					if (info.remainingQuantity < remainLimit)
					{
						well.setError('Not enough DNA');
					}
					// Warning if near the limit
					if (info.remainingQuantity < LOW_QUANTITY_WARNING_LIMIT && !info.preNormalized)
					{
						warningMsg[warningMsg.length] = 'Low quantity';
					}
*/
/*
					if (quantityLeft < 1000.0*remainLimit)
					{
						well.setError('Not enough DNA');
					}
					// Warning if near the limit
					if (quantityLeft < 1000.0*LOW_QUANTITY_WARNING_LIMIT && !info.preNormalized)
					{
						warningMsg[warningMsg.length] = 'Low quantity';
					}
*/
/*
				}
				else
				{
					warningMsg[warningMsg.length] = 'No quantity';
				}
*/
				// --- Used volume+water
				if (info.origQubitConc != null)
				{
					text += '<div class="volumes"><span class="volume">'+Numbers.formatNumber(volDNA, 2)+'</span> + <span class="water">'+Numbers.formatNumber(water, 2)+'µl</span></div>';
				}
				// --- ΔCt
				if (info.origDeltaCt != null)
				{
					text += '<div class="quality-score">ΔCt='+Numbers.formatNumber(info.origDeltaCt, 2) + '</div>';
					if (info.origDeltaCt > QUALITY_SCORE_WARNING_LIMIT)
					{
						if (!painter.hasClass('plate', 'hide-quality-score'))
						{
							warningMsg[warningMsg.length] = 'High ΔCt value';
						}
					}
				}
				else
				{
					if (!painter.hasClass('plate', 'hide-quality-score'))
					{
						warningMsg[warningMsg.length] = 'No ΔCt value';
					}
				}
				// --- QubitConc
				if (info.origQubitConc)
				{
					text += '<div class="origqubitconc">'+Numbers.formatNumber(info.origQubitConc, 2) + 'ng/µl (orig.)</div>';
					text += '<div class="calcqubitconc">'+Numbers.formatNumber(calcConc, 2) + 'ng/µl (calc.)</div>';
				}
				else
				{
					if (!painter.hasClass('plate', 'hide-origqubitconc'))
					{
						warningMsg[warningMsg.length] = 'No QubitConc value';
					}
				}
/*
				if (info.QiacubeDate)
				{
					text += '<div class="qiacube-date">'+info.QiacubeDate+'</div>';
				}
				else if (info.DilutionDate)
				{
					text += '<div class="dilution-date">'+info.DilutionDate+'</div>';
				}
				else
				{
					warningMsg[warningMsg.length] = info.preNormalized ? 'No DilutionDate value' : 'No QiacubeDate value';
				}
*/
				if (info.AutoProcessing)
				{
					warningMsg[warningMsg.length] = info.AutoProcessing;
				}
/*
				dna.usedQuantity = dna.qc ? QUANTITY_QC : QUANTITY_REGULAR; // µg
				dna.dilutionConc = 1000 * dna.usedQuantity / TOTAL_VOLUME; // ng/µl

				var volDNA = Math.ceil(10000*dna.usedQuantity/info.origQubitConc) / 10; // µl, rounded to 1 decimal
				var water = TOTAL_VOLUME - volDNA;
				
				if (volDNA < MINIMAL_DNA_VOLUME)
				{
					// Large mix since we do not want to take less than 1µl
					volDNA = MINIMAL_DNA_VOLUME;
					dna.usedQuantity = volDNA * info.origQubitConc / 1000; // µg
					var totalVolume = 1000 * dna.usedQuantity / dna.dilutionConc; // µl
					water = totalVolume - volDNA;
					warningMsg[warningMsg.length] = 'Large mix';
				}
				
				if (info.remainingQuantity)
				{
					//text += '<div class="quantity">'+Numbers.formatNumber(info.remainingQuantity, 2) + 'µg</div>';
					text += '<div class="quantity">'+Numbers.formatNumber(info.remainingQuantity, 2) + 'µg '+Numbers.formatNumber(volDNA, 1)+'µl</div>';
					// Must have at least 1.1µg or 1.22µg (if default values are used)
					var remainLimit = dna.qc ? QUANTITY_QC : QUANTITY_REGULAR;
					if (info.remainingQuantity < remainLimit)
					{
						well.setError('Not enough DNA');
					}
					// Warning if near the limit
					if (info.remainingQuantity < LOW_QUANTITY_WARNING_LIMIT && !info.preNormalized)
					{
						warningMsg[warningMsg.length] = 'Low quantity';
					}
				}
				else
				{
					warningMsg[warningMsg.length] = 'No quantity';
				}
				if (info.origDeltaCt)
				{
					text += '<div class="quality-score">ΔCt='+Numbers.formatNumber(info.origDeltaCt, 1) + '</div>';
					if (info.origDeltaCt > QUALITY_SCORE_WARNING_LIMIT) warningMsg[warningMsg.length] = 'High ΔCt value';
				}
				else
				{
					warningMsg[warningMsg.length] = 'No ΔCt value';
				}
				if (info.origQubitConc)
				{
					text += '<div class="origqubitconc">'+Numbers.formatNumber(info.origQubitConc, 2) + 'ng/µl</div>';
					text += '<div class="volumes"><span class="volume">'+Numbers.formatNumber(volDNA, 1)+'</span> + <span class="water">'+Numbers.formatNumber(water, 1)+'µl</span></div>';
				}
				else
				{
					warningMsg[warningMsg.length] = 'No QubitConc value';
				}
				if (info.QiacubeDate)
				{
					text += '<div class="qiacube-date">'+info.QiacubeDate+'</div>';
				}
				else if (info.DilutionDate)
				{
					text += '<div class="dilution-date">'+info.DilutionDate+'</div>';
				}
				else
				{
					warningMsg[warningMsg.length] = info.preNormalized ? 'No DilutionDate value' : 'No QiacubeDate value';
				}
				if (info.AutoProcessing)
				{
					warningMsg[warningMsg.length] = info.AutoProcessing;
				}
*/
			}
			else if (!dna.origId)
			{
				well.setError('DNA not found');
			}
			if (dna.comment)
			{
				text += '<div class="comment">'+Strings.encodeTags(dna.comment)+'</div>';
			}
			if (info && info.comment)
			{
				text += '<div class="comment">'+Strings.encodeTags(info.comment)+'</div>';
			}
			if (warningMsg.length > 0)
			{
				well.setWarning(warningMsg.join('; '));
			}
			// Index 1 & 2 primers
			var row = well.row + 1;
			var column = well.column + 1;
			var index2 = 'A50' + row;
			var index1 = 'A70' + column;
			if (column > 9)
			{
				index1 = 'A7' + column;
			}
			text += '<div class="indexprimer">'+index2+'</div>';
			text += '<div class="indexprimer">'+index1+'</div>';
		}
		else if (well.copyText)
		{
			text = '<div class="copy-text">'+well.copyText+'</div>';
		}
		else
		{
			text = 'empty';
		}
		if (well.duplicate)
		{
			well.setError('Duplicate barcode');
		}

		var bc = Doc.element('barcode.'+well.row+'.'+well.column);
		if (bc)
		{
			var cls = 'barcode-well'
			cls += well.column == 0 ? ' barcode-left' : ' barcode-right';

			if (well.barcode)
			{
				bc.innerHTML = well.barcode.name;
				var indexSet = painter.barcodeVariant ? painter.barcodeVariant.indexSets[well.column] : null;
				if (indexSet)
				{
					cls += ' ' + indexSet.color;
				}
				if (well.defaultBarcode && well.barcode != well.defaultBarcode)
				{
					cls += ' bg-modified';
				}
				if (well.duplicate)
				{
					cls += ' duplicate';
				}
			}
			bc.className = cls;
		}
		return text;
	}

	/**
	 * Returns `true` if HTML element with id 'elementId' has
	 * class 'cls', else `false`.
	 */
	painter.hasClass = function(elementId, cls)
	{
		var element = document.getElementById(elementId);
		return (' ' + element.className + ' ').indexOf(' ' + cls + ' ') > -1;
	}

	painter.fetchDilutionFactor = function(deltaCt)
	{
		var dilutionFactor = 1.0;
/*
		// Illumina ladder dilution chart
		if (deltaCt > 0.5 && deltaCt <= 1.5)
		{
			dilutionFactor = 2.0;
		}
		if (deltaCt > -0.5 && deltaCt <= 0.5)
		{
			dilutionFactor = 4.0;
		}
		if (deltaCt > -1.5 && deltaCt <= -0.5)
		{
			dilutionFactor = 8.0;
		}
		if (deltaCt <=- 1.5)
		{
			dilutionFactor = 16.0;
		}
*/
		// Dilution curve fitted to centers of ladder steps in Illumina ladder dilution chart (Mats Jönsson 2015-09-28)
		var fact = 2.8284271247;
		var expFact = 0.6931471806;
		dilutionFactor = fact * Math.exp(-expFact*deltaCt);
		// Apply cut-offs restricting the dilution factor to [1.0, 16.0]
		if (dilutionFactor < 1.0)
		{
			dilutionFactor = 1.0;
		}
		if (dilutionFactor > 16.0)
		{
			dilutionFactor = 16.0;
		}
		return dilutionFactor;
	}
	
	return painter;
}();
